from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import str

import sys
import string
import json
import logging

from .colors import *


def is_posix_filename(name, extra_chars=""):
    CHARS = string.ascii_letters + string.digits + "._-" + extra_chars
    return all(c in CHARS for c in name)


def cprint(color, msg, file=sys.stderr, end='\n'):
    if type(msg) is str:
        data = msg
    # 2to3: elif type(msg) is str:
    elif isinstance(msg, str):
        data = msg.__str__()
    else:
        data = str(msg)

    print(color + data + RESET, file=file, end=end)


def _assert(condition, msg):
    if condition:
        return

    cprint(LRED, msg)
    logging.error(msg)
    sys.exit(1)


def to_unicode(obj, encoding='utf-8'):
    # 2to3: assert isinstance(obj, basestring), type(obj)
    #assert isinstance(obj, str), type(obj)
    if isinstance(obj, str):
        return obj

    for encoding in ['utf-8', 'latin1']:
        try:
            obj = str(obj, encoding)
            return obj
        except UnicodeDecodeError:
            pass

    assert False, "tst: non-recognized encoding"


def data2json(data):
    def date_handler(obj):
        if hasattr(obj, 'isoformat'):
            return obj.isoformat()
        elif hasattr(obj, 'email'):
            return obj.email()

        return obj

    return json.dumps(
        data,
        default=date_handler,
        indent=2,
        separators=(',', ': '),
        sort_keys=True,
        ensure_ascii=False)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import hashlib
import io
import os
import sys

import tst
from tst.utils import data2json, _assert, cprint
from tst.colors import *
from tst.jsonfile import JsonFile

def main():
    _assert(len(sys.argv) in [3, 4], 'Usage: tst commit [--support-file|-s] <filename>')
    _assert(os.path.exists('.tst/assignment.json'), 'No tst assignment found')
    filetype = 'support' if sys.argv[2] in ['--support-file', '-s'] else 'answer'
    filename = sys.argv[3 if filetype == 'support' else 2]
    assignment_id = JsonFile('.tst/assignment.json')['iid']
    site = tst.get_site('_DEFAULT')
    commit(filename, filetype, assignment_id, site)


def read_mode(filename):
    return "rw"


def read_content(filename):
    with io.open(filename, encoding='utf-8') as f:
        content = f.read()

    return content


def commit(filename, filetype, key, site):
    content = read_content(filename)
    checksum = hashlib.sha1(content.encode('utf-8')).hexdigest()
    data = {
        "files": [{
                "name": filename,
                "content": content,
                "mode": read_mode(filename),
                "category": "public",
                "type": filetype,
                "hash": checksum
            }],
        "hash": checksum
    }

    response = site.send_answer(data, key)
    if response.ok:
        cprint(LGREEN, 'File saved successfully')
    elif response.status_code == 401:
        cprint(LRED, 'Commit failed (%s)' % response.status_code)
        cprint(LRED, 'run: tst login')
    elif response.status_code == 412:
        cprint(LRED, 'Commit failed (%s)' % response.status_code)
        cprint(LRED, 'Error: %s' % response.json()['messages'][0])
    else:
        cprint(LRED, 'Commit failed (%s)' % response.status_code)
        cprint(LRED, 'Error: %s' % response.json()['messages'][0])



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import str

import sys
import os
import io
import json
import glob
import datetime as dt
import pkg_resources
import logging

import requests
from cachecontrol import CacheControl
from cachecontrol.caches.file_cache import FileCache

from .jsonfile import JsonFile, CorruptedJsonFile
from .colors import *
from .utils import cprint, _assert, is_posix_filename, data2json

CONFIGDIR = os.path.expanduser('~/.tst/')
LOG_FILE = os.path.expanduser('~/.tst/logs.txt')
CONFIGFILE = CONFIGDIR + 'config.yaml'

def coverit():
    return 1

def get_config():
    if not os.path.exists(CONFIGFILE):
        if not os.path.isdir(CONFIGDIR):
            os.mkdir(CONFIGDIR)

        with io.open(CONFIGFILE, encoding="utf-8", mode="w") as config_file:
            config_file.write(
                "run:\n"
                "  py: python3\n"
                "  py2: python2\n\n"
                "sites:\n"
                "- name: demo\n"
                "  url: https://raw.githubusercontent.com/daltonserey/tst-demo/master\n"
            )

    if not os.path.exists(LOG_FILE):
        open(LOG_FILE, 'w').close()
    logging.basicConfig(filename=LOG_FILE, level=logging.DEBUG, format='%(asctime)s (%(levelname)s@%(name)s) %(message)s', datefmt='%Y-%m-%dT%H:%M:%S')
    # Uncomment to reduce too much data in logs
    #logging.getLogger("requests").setLevel(logging.WARNING)
    #logging.getLogger("cachecontrol.controller").setLevel(logging.WARNING)
    #logging.getLogger("urllib3.connectionpool").setLevel(logging.WARNING)

    return JsonFile(CONFIGFILE)


def dirtype(path=""):
    path = os.path.abspath(os.path.expanduser(path))

    # tst internal types of directories
    if path == os.path.expanduser('~/.tst'):
        return "config"

    elif os.path.basename(path) == '.tst':
        return "internal"

    # user content
    elif os.path.exists(path + '/.tst/assignment.json'):
        return "assignment"

    elif os.path.exists(path + '/.tst/activity.json'):
        return "activity"

    elif os.path.exists(path + '/.tst/collection.json'):
        return "collection"

    # user content (old format)
    elif os.path.exists(path + '/.tst/tst.json'):
        kind = JsonFile(path + "/.tst/tst.json").get("kind", "")
        return "old:" + kind

    # corrupted/incomplete content
    elif os.path.isdir(path + '/.tst') and not path == os.path.expanduser('~'):
        return "corrupted"

    # directory contains a file with tst tests
    elif os.path.exists(path + '/tst.json') or os.path.exists(path + '/tst.yaml'):
        return "tst-able"

    # not a tst directory
    return None


def validate_tst_object(json):
    def is_valid_mode(mode):
        # 2to3: isinstance(mode, basestring) and\
        return mode is None or\
               isinstance(mode, str) and\
               len(mode) <= 3 and\
               all(d in 'rwxo' for d in mode.lower())

    assert 'files' in json, "missing files property"
    assert type(json) is dict, "json is not an object"
    assert 'kind' in json, "missing kind property"
    assert json['kind'] in ["assignment", "activity", "collection"], "unrecognized kind"

    files = json['files']
    assert all('content' in f for f in files), "missing content in file(s)"
    assert all('name' in f for f in files), "missing name in file(s)"
    assert all(is_posix_filename(f['name'], "/") for f in files), "non posix portable file(s) name(s)"
    assert all(is_valid_mode(f.get('mode')) for f in files), "invalid mode in file(s)"
    #assert len(files) == len(set([f['name'] for f in files])), "repeated file names"
    if not (len(files) == len(set([f['name'] for f in files]))):
        filenames = [f['name'] for f in files]
        repeated = list(set([fn for fn in filenames if filenames.count(fn) > 1]))
        msg = f"file repeated: {' '.join(repeated)}"
        logging.warning(msg)


def save_file(filename, content, mode):
    def octal_mode(mode):
        return {
            (False , False): 0o444,
            (False ,  True): 0o555,
            (True  , False): 0o644,
            (True  ,  True): 0o755
        }["w" in mode, "x" in mode]

    subdirs = os.path.dirname(filename)
    if not os.path.isdir(subdirs):
        os.makedirs(subdirs)

    with io.open(filename, encoding="utf-8", mode="w") as f:
        f.write(content)

    os.chmod(filename, octal_mode(mode))


def save_files(files, basedir, verbose=True):
    saved = 0
    for f in files:
        filename = basedir + "/" + f['name']
        mode = f.get('mode', '644')

        try:
            if os.path.exists(filename):
                os.chmod(filename, 0o644)

            verbose and cprint(LGREEN, "W %s" % filename)
            save_file(filename, f['content'], mode)
            saved += 1

        except IOError as e:
            cprint(LRED, e)
            cprint(LRED, "Failed saving file: '%s'" % f['name'])

        except OSError as e:
            cprint(YELLOW, e)
            cprint(LRED, "Failed setting file mode: '%s'" % f['name'])

    return saved


def parse_file_spec(fspec):

    # normalize fspec
    while fspec.count(',') < 2:
        fspec = fspec + ','

    # separate filename and details
    data = fspec.split(",")
    filename = data[0].strip()
    details = [e.strip() for e in data[1:]]

    # consume public or private specification
    if 'public' in details:
        public = True
        details.remove('public')

    elif 'private' in details:
        public = False
        details.remove('private')

    else:
        public = False
        details.remove('')

    # check mode
    mode = details[0]

    return filename, 'public' if public else 'private', mode


def is_single_line_string(text):
    return '\n' not in text and '\r' not in text


class Site:
    def __init__(self, name=None, url=None):
        self.name = name
        self.url = url
        self.last_error = None
        self._session = None
        self._googappuid = None
        sites = get_config()['sites']
        for s in sites:
            if s['name'] == name:
                self.url = s['url']


    def api_url(self):
        urls = self.urls()
        assert urls, "no urls provided by site"
        url = self.api_full_url(urls.get('api-url'))
        if url.startswith("http://https"):
            url = url[7:]
        return url


    def api_login_url(self):
        urls = self.urls()
        if urls is None:
            return None

        return self.api_full_url(urls.get('login-url'))


    def api_access_url(self):
        urls = self.urls()
        if urls is None:
            return None

        return self.api_full_url(urls.get('access-url'))


    def api_full_url(self, url):
        if url is None:
            return None

        if url.startswith('http://') or url.startswith('https://'):
            return url

        if url.startswith('/'):
            return f"{self.urls()['api-url']}{url}"

        return f"{self.urls['api-url']}/{url}"


    def login_url(self):
        urls = self.urls()
        if urls is None:
            return None

        return self.full_url(urls.get('login'))


    def token_url(self):
        urls = self.urls()
        if urls is None:
            return None

        return self.full_url(urls['token'])


    def full_url(self, url):
        if url is None:
            return None

        if url.startswith('http://') or url.startswith('https://'):
            return url

        if url.startswith('/'):
            return self.url + url

        return self.url + '/' + url


    def urls(self):
        s = self.get_session()
        try:
            response = s.get(self.url + '/tst.json', allow_redirects=True)
        except requests.ConnectionError:
            _assert(False, "Connection failed... check your internet connection")

        if not response.ok:
            return None

        response.encoding = 'utf-8'
        try:
            resource = response.json()
            resource['_response'] = response

        except ValueError:
            return None

        return resource


    def save_cookies(self, response):
        cookies_file = JsonFile(os.path.expanduser('~/.tst/cookies.json'))
        cookies_file.writable = True
        new_googappuid = response.cookies['GOOGAPPUID'] if response.cookies else None
        if not new_googappuid:
            logging.warning("no GOOGAPPUID cookie in server response (setting value to 0)")
            new_googappuid = '0'

        if not new_googappuid.isdigit() or not (0 <= int(new_googappuid) < 1000):
            logging.warning("invalid value for GOOGAPPUID cookie: %s (setting value to 0)" % new_googappuid)
            new_googappuid = '0'

        if new_googappuid != self._googappuid:
            logging.info("saving new GOOGAPPUID cookie: %s" % new_googappuid)
            cookies_file[self.name] = {"GOOGAPPUID": new_googappuid}
            cookies_file.save()

    def get(self, path):
        s = self.get_session()
        url = "%s%s" % (self.api_url(), path) if path.startswith('/') else path
        try:
            response = s.get(url, allow_redirects=True)
            logging.info('GET %s (%s)' % (url, response.status_code))
        except requests.ConnectionError:
            _assert(False, "Connection failed... check your internet connection")

        if not response.ok:
            self.last_error = response.status_code
            self.last_response = response
            return response

        self.save_cookies(response)
        response.encoding = 'utf-8'
        return response



    def get_activity(self, key):
        s = self.get_session()
        url = "%s/%s" % (self.api_url(), key)
        #url = url.replace('api.','dev.api.')
        #cprint(YELLOW, url)
        try:
            response = s.get(url, allow_redirects=True)
            logging.info('GET %s (%s)' % (url, response.status_code))
        except requests.ConnectionError:
            _assert(False, "Connection failed... check your internet connection")

        if not response.ok:
            self.last_error = response.status_code
            self.last_response = response
            return (None, response)

        self.save_cookies(response)
        response.encoding = 'utf-8'
        try:
            resource = response.json()
            resource['_response'] = response
            validate_tst_object(resource)

        except ValueError:
            #_assert(False, "Resource is not valid json")
            logging.warning('ValueError during resource processing (in get_activity())')
            return (None, response)

        except AssertionError as e:
            _assert(False, "Not a TST Object: %s" % e.message)

        return (resource, response)

    def get_directory(self, key):
        s = self.get_session()
        api_url = self.api_url()
        url = "%s/%s/tst.yaml" % (api_url, key)
        try:
            response = s.get(url, allow_redirects=True)
            logging.info('GET %s (%s)' % (url, response.status_code))
        except requests.ConnectionError:
            _assert(False, "Connection failed... check your internet connection (1)")

        # process response
        if not response.ok:
            self.last_error = response.status_code
            self.last_response = response
            return (None, response)

        response.encoding = 'utf-8'
        try:
            import yaml
            resource = yaml.load(response.text, Loader=yaml.FullLoader)
            resource['_response'] = response

        except Exception as e:
            cprint(YELLOW, "Activity malformed: %s" % url)
            logging.info('Falied parsing activity yaml' % (url, response.status_code))
            raise e

        # gather files
        files = resource.get('files') or []
        files.append({
            "name": "tst.yaml",
            "content": response.text,
            "mode": "ro"
        })

        ## add text file if required
        if 'text' in resource and is_single_line_string(resource['text']):
            files.append({
                "name": resource['text'],
                "content": '%s/%s/%s' % (api_url, key, resource['text']),
                "mode": "ro"
            })

        ## add included files
        files_filenames = [f['name'] for f in files]
        for fspec in resource['include']:
            filename, category, mode = parse_file_spec(fspec)
            if filename not in files_filenames:
                files.append({
                    'name': filename,
                    'content': '%s/%s/%s' % (api_url, key, filename),
                    'mode': mode
                })
            else:
                entry = next(e for e in files if e['name'] == filename)
                entry['mode'] = mode

        ## fetch missing files
        for f in files:
            if f['content'].startswith('http://') or f['content'].startswith('https://'):
                url = '%s/%s/%s' % (api_url, key, f['name'])
                try:
                    response = s.get(url)
                    logging.info('GET %s (%s)' % (url, response.status_code))
                except requests.ConnectionError:
                    _assert(False, "Connection failed... check your internet connection")

                _assert(response.ok, "%s\nFile request failed: %s (%d)" % (url, response.reason, response.status_code))
                response.encoding = 'utf-8'
                f['content'] = response.text

        return ({
            'kind': 'activity',
            'files': files,
        }, response)


    def get_session(self, headers=None, cookies=None):
        """
        Return a pre-configured requests session object, containing
        the proper headers, cookies and cache control decorator.
        """
        # is there a previous session?
        if self._session:
            return self._session

        # create basic session objet
        s = requests.session()

        # add headers
        s.headers = headers or { 'X-TST-Version': pkg_resources.get_distribution('tst').version }
        token = JsonFile(os.path.expanduser('~/.tst/tokens.json')).get(self.name)
        if token:
            s.headers['Authorization'] = 'Bearer %s' % token

        # set cookies
        allcookies = JsonFile(os.path.expanduser('~/.tst/cookies.json'))
        cookies = cookies or allcookies.get(self.name) or {}
        if cookies:
            self._googappuid = cookies['GOOGAPPUID']
            logging.info("setting up session with GOOGAPPUID cookie: %s" % self._googappuid)
        else:
            logging.warning("setting up session WITHOUT GOOGAPPUID")
        s.cookies.update(cookies)

        self._session = CacheControl(s, cache=FileCache(os.path.expanduser('~/.tst/cache')))
        return self._session

    def send_answer(self, answer, key):
        s = self.get_session()
        url = "%s/%s/answers" % (self.api_url(), key)
        #url = url.replace('api.', 'dev.api.')
        #cprint(YELLOW, url)
        data = data2json(answer).encode('utf-8')
        try:
            response = s.post(url, data=data, allow_redirects=True)
        except requests.ConnectionError:
            _assert(False, "Connection failed... check your internet connection (1)")

        return response

    def post(self, target, data, headers=None, cookies=None):
        # BEWARE: this is an attempt to make a simple facility method
        #         configures and performs an HTTP POST request to
        #         the given site.
        #
        # target: is either a target within the site or the full url
        s = self.get_session(headers=headers, cookies=cookies)
        url = "%s%s" % (self.url, target) if target.startswith('/') else target
        response = s.post(url, data=data2json(data), allow_redirects=True)
        logging.info('POST %s (%s)' % (url, response.status_code))
        self.save_cookies(response)
        return response


def get_site(name=None, url=None):
    assert name is None or url is None

    if url:
        return Site(url=url)

    elif name == "_DEFAULT":
        index = 0

    else:
        sites = get_config()['sites']
        index = next((i for i in range(len(sites)) if sites[i]['name'] == name), None)
        if index is None:
            return None

    return Site(name=get_config()['sites'][index]['name'])


def read_specification(filename=None, verbose=False):
    # deal with a custom specification file name
    if filename:
        _assert(os.path.exists(filename), "File %s not found" % filename)
        cprint(LCYAN, "Reading specification file: %s" % filename)
        return JsonFile(filename, array2map="tests")

    # deal with default specification file names
    tstyaml_exists = os.path.exists('tst.yaml')
    tstjson_exists = os.path.exists('tst.json')
    if verbose and tstyaml_exists and tstjson_exists:
        cprint(YELLOW, "Found both tst.yaml and tst.json: using tst.yaml")

    if tstyaml_exists:
        try:
            specification = JsonFile('tst.yaml', array2map="tests")
        except CorruptedJsonFile:
            _assert(False, "Corrupted specification file")
        return specification

    elif tstjson_exists:
        try:
            specification = JsonFile('tst.json', array2map="tests")
        except CorruptedJsonFile:
            _assert(False, "Corrupted specification file")
        return specification

    # neither tst.yaml, nor tst.json exist
    candidates = glob.glob("*.yaml")
    if len(candidates) == 0:
        candidates = glob.glob("*.json")

    if len(candidates) == 1:
        cprint(YELLOW, "Using %s as specification file" % candidates[0])
        try:
            specification = JsonFile(candidates[0], array2map="tests")
        except:
            _assert(False, "Invalid specification file")
        return specification

    cprint(LRED, "No tst tests found")
    sys.exit(1)


def save_assignment(activity, dir_name, etag, url, repo):

    # move into directory
    os.chdir(dir_name)

    # save the original activity data
    dirname = './.tst'
    if not os.path.exists(dirname):
        os.makedirs(dirname)

    with io.open('./.tst/activity.json', mode='w', encoding='utf-8') as f:
        f.write(data2json({
            "url": url,
            "name": activity.get('name'),
            "activity": activity,
            "etag": etag,
            "repo": repo,
            "updated_at": dt.datetime.utcnow().isoformat().split(".").pop(0) + "Z"
        }))

    # save activity files
    files = activity['files']
    for file in files:
        if os.path.exists(file['name']):
            contents = open(file['name']).read().decode('utf-8')
            if contents != file['data']:
                cprint(LRED, "skipping modified file: '%s' (use --overwrite)" % file['name'])
            else:
                cprint(RESET, "skipping unmodified file: '%s'" % file['name'])
            continue

        try:
            with io.open(file['name'], mode='w', encoding='utf-8') as f:
                f.write(file['data'])
            cprint(LCYAN, "Adding file '%s'" % file['name'])
        except:
            print("tst: fatal: Can't save file '%s'" % file['name'], file=sys.stderr)
            sys.exit(1)



# coding: utf-8
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import input, str

import sys
import curses
from curses import wrapper

def find_getch():
    try:
        import termios
    except ImportError:
        # Non-POSIX. Return msvcrt's (Windows') getch.
        import msvcrt
        return msvcrt.getch

    # POSIX system. Create and return a getch that manipulates the tty.
    import tty
    def _getch():
        fd = sys.stdin.fileno()
        old_settings = termios.tcgetattr(fd)
        try:
            tty.setraw(fd)
            ch = sys.stdin.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
        return ch

    return _getch


def climenu(options, try_curses=False, prompt=None):
    if type(options) is dict:
        items = [(str(k), options[k]) for k in options]
        original_keys = {str(k): k for k in options.keys()}
    else:
        items = [(str(i + 1), options[i]) for i in range(len(options))]
        original_keys = {str(i + 1): i for i in range(len(options))}

    single_char_keys = all(len(str(k)) == 1 for k, text in items)
    items.sort(key=lambda e: "%05d" % int(e[0]) if str(e[0]).isdigit() else e[0])
    if try_curses and single_char_keys:
        selected = wrapper(menu_curses, items, prompt)
    else:
        selected = menu_term(items, prompt)

    return original_keys[selected]


def menu_term(items, prompt=None):
    for key, text in items:
        print('%s: %s' % (key, text), file=sys.stderr)

    options_keys = [key for key, text in items]
    single_char_keys = all(len(k) == 1 for k in options_keys)
    read_input = find_getch() if single_char_keys else input
    while True:
        print(prompt or '', end='')
        key = read_input()
        if key in options_keys: break

    return key


def menu_curses(stdscr, items, prompt):
    stdscr.clear()
    h, w = stdscr.getmaxyx()
    assert h > len(items), "oops… screen is too small for menu"
    first_line = h - len(items) - (1 if prompt else 0)
    for i in range(len(items)):
        key = items[i][0]
        value = items[i][1]
        stdscr.addstr(first_line + i, 0, "%s. %s" % (key, value), curses.A_BOLD)

    keys_codes = [ord(key) for key, text in items]
    while True:
        if prompt:
            stdscr.addstr(h-1, 0, prompt)
            stdscr.refresh()
        key_code = stdscr.getch()
        if key_code in keys_codes: break

    return chr(key_code)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import input

import tst
import sys
import os

from urllib.parse import urlparse

from tst.colors import *
from tst.utils import cprint
from tst.utils import _assert
from tst.utils import is_posix_filename
from tst.utils import data2json
from tst.jsonfile import JsonFile

def main():
    assert sys.argv[0].endswith('tst') and sys.argv[1] == 'checkout'

    # parse user arguments
    args = sys.argv[2:]
    site, key, target_dir, overwrite = process_args(args)

    checkout(site, key, target_dir, overwrite)


def syntax_help():
    return LRED + "This is not a tst directory.\n" + RESET +\
           "Usage: tst checkout <key>[@<site>] [<directory>]\n"\
           "       tst checkout <url> [<directory>]\n\n"\
           "   or: tst checkout\n"\
           "       (inside a tst directory)"\


def process_args(args):
    _assert(0 <= len(args) <= 3, syntax_help())

    # process overwrite argument
    YES = ["--yes", "-y"]
    overwrite = any(yes in args for yes in YES)
    args = [e for e in args if e not in YES]

    # add default args
    if len(args) == 0:
        ## no args: use current directory
        cwd = os.getcwd()
        _assert(tst.dirtype(cwd) == "assignment", "This directory is not a tst activity")
        args.append(os.path.basename(cwd))
        args.append(".")

    elif len(args) == 1 and os.path.isdir(args[0]):
        ## single directory as arg: make sure is a checkout
        checkout_file = args[0] + "/.tst/assignment.json"
        _assert(os.path.exists(checkout_file), "Directory %s has no checkout info\n%sTry: tst checkout key <new-directory>" % (args[0], RESET))
        tst_object = JsonFile(checkout_file)
        _assert("key" in tst_object, "Checkout corrupted: missing key")
        _assert("site" in tst_object, "Checkout corrupted: missing site")
        args = ["%s@%s" % (tst_object["key"], tst_object["site"]), args[0]]

    data = { "destdir": args[1] if len(args) > 1 else None }

    key_or_url = args[0]
    url = urlparse(key_or_url)
    _assert(not url.fragment, "Fragments not supported in url")
    _assert(not url.username, "Usernames not supported in url")
    _assert(not url.password, "Passwords not supported in url")

    if url.scheme == 'http':
        # key_or_url == "http://.../key"
        path, key = url.path.rsplit("/", 1) if "/" in url.path else (url.path, '')
        data["url"] = "%s://%s%s" % (url.scheme, url.netloc, path)
        data["site"] = None
        data["key"] = key

    elif is_posix_filename(key_or_url, extra_chars="@"):
        num_ats = key_or_url.count("@")
        if num_ats == 0:
            # key_or_url == "key"
            data["url"] = None
            data["site"] = "_DEFAULT"
            data["key"] = key_or_url

        elif num_ats == 1:
            # key_or_url == "key@site"
            key, sitename = key_or_url.split("@")
            data["url"] = None
            data["site"] = sitename
            data["key"] = key
    else:
        _assert(False, "Unrecognized key/url")


    # get site based either on sitename or url
    assert data['site'] is None or data['url'] is None
    site = tst.get_site(name=data['site']) if data['site'] else tst.get_site(url=url)
    _assert(site and site.url, "No site/url identified. Check your config.yaml")

    return site, data['key'], data['destdir'], overwrite


def existing_files(basedir, files):
    filenames = []
    for f in files:
        fullname = "%s/%s" % (basedir, f['name'])
        if os.path.exists(fullname):
            filenames.append(fullname)

    return filenames


def checkout(site, key, target_dir, overwrite):
    """checkout activity/assignment from site/collection"""

    def is_valid_dir(dirtype):
        return dirtype in [None, "assignment"]

    # fetch activity
    cprint(LGREEN, "Fetching %s from %s" % (key, site.name or site.url))
    activity, response = site.get_activity(key)
    if response.status_code == 404:
        activity, response = site.get_directory(key)

    if response.status_code == 401:
        cprint(LRED, 'Checkout failed (%s)' % site.last_response.status_code)
        cprint(WHITE, 'run: tst login')
        return

    elif response.status_code == 404:
        cprint(LRED, 'Checkout failed (%s)' % site.last_response.status_code)
        cprint(LRED, 'Assignment not found')
        return

    elif response.status_code == 412:
        cprint(LRED, 'Checkout failed (%s)' % site.last_response.status_code)
        cprint(LRED, 'Error: %s' % site.last_response.json()['messages'][0])
        return

    _assert(activity, "CANNOT checkout %s from site %s (%s)" % (key, site.name, response.status_code))

    # set destination directory
    destdir = target_dir or activity.get('dirname') or activity.get('name') or key
    if not target_dir:
        cprint(YELLOW, "Directory argument not found")
        cprint(RESET, "(You can add directory as an additional argument)")
        while True:
            destdir = input("Provide directory name (default %s): " % destdir) or key
            if is_posix_filename(destdir, extra_chars="/"): break
            cprint(YELLOW, "Invalid portable posix filename: '%s'" % destdir)

    _assert(not os.path.exists(destdir) or (os.path.isdir(destdir) and is_valid_dir(tst.dirtype(destdir))), "Invalid target directory: %s" % destdir)

    # check whether files exist
    old_files = existing_files(destdir, activity['files'])
    if old_files and not overwrite:
        cprint(YELLOW, "If you proceed, these files will be overwriten")
        for fn in old_files:
            cprint(LCYAN, fn)

        cprint(YELLOW, "Proceed (y/N)? ", end="")
        if input() != "y":
            cprint(LRED, 'Aborting checkout')
            sys.exit(1)

    # save files
    saved = tst.save_files(activity['files'], destdir)
    cprint(LBLUE, "%d files saved to %s%s%s directory" % (saved, LCYAN, destdir, LBLUE))
    if len(activity['files']) > saved:
        cprint(YELLOW, "%d files were NOT saved" % (len(activity['files']) - saved))

    internal = [{
        "name": ".tst/assignment.json",
        "content": data2json({
            "kind": "assignment",
            "site": site.name,
            "key": key,
            "iid": activity.get('iid'),
            "user": activity.get('user'),
        })
    }]
    tst.save_files(internal, destdir, verbose=False)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import str

import sys
import signal
import json

from subprocess import Popen, PIPE, CalledProcessError

import tst
from colors import *
from utils import to_unicode, cprint

class ConnectionFail(Exception): pass

class Server(object):
    
    __instance = None

    class Response:

        def json(self):
            if '_json' not in dir(self):
                try:
                    self._json = json.loads(self.body)
                except:
                    self._json = None

            return self._json


    def __new__(cls):

        # instantiation
        if Server.__instance is not None:
            return Server.__instance

        Server.__instance = object.__new__(cls)
        self = Server.__instance

        # initialization
        #self.config = Config()
        self.config = tst.get_config()
        self.user = self.config.get('user')
        self.token = self.config.get('access_token')

        return self


    def request(self, method, path, headers={}, payload=None, exit_on_fail=False):
        curl_command = [
            'curl',
            '-q', # don't use ~/.curlrc (must be first arg)
            '-X', method.upper(), # http verb
            '-v', # be verbose: print report to stderr
            '-s', # don't print progress meter
            '-L'  # follow redirects
        ]

        headers['TST-CLI-Release'] = self.config.get('release', 'unknown')
        if 'Authorization' not in headers:
            headers['Authorization'] = 'Bearer %s' % self.token
        for hname, hvalue in headers.items():
            curl_command.append('-H')
            curl_command.append('%s: %s' % (hname, hvalue))

        url = self.config['url'] + path
        curl_command.append(url)
        if payload is not None:
            curl_command.append('-d')
            data = "%s" % json.dumps(payload)
            curl_command.append(data)

        signal.alarm(20000) # timeout in seconds
        process = Popen(curl_command, stdout=PIPE, stderr=PIPE) 
        try:
            stdout, stderr = map(to_unicode, process.communicate())
            signal.alarm(0) # reset alarm for future use...
            process.wait()
        except: # timeout!!!
            process.terminate()
            raise

        # raw data
        response = self.Response()
        response.stderr = stderr
        response.stdout = stdout
        response.exit_status = process.returncode

        # curl messages
        lines = [l[2:] for l in stderr.splitlines() if l and l[0] == '*']
        response.curl_messages = "\n".join(lines)

        # request headers
        lines = [l[2:] for l in stderr.splitlines() if l and l[0] == '>']
        response.request_headers = "\n".join(lines)

        # response headers
        lines = [l[2:] for l in stderr.splitlines() if l and l[0] == '<']
        response.headers = "\n".join(lines)

        if not response.headers:
            if exit_on_fail:
                msg = "tst: can't connect to server"
                _assert(False, msg)
                
            raise ConnectionFail("can't connect to tst online")

        # body
        response_lines = response.headers.splitlines()
        response.status_code = None
        for i in range(len(response_lines)-1, -1, -1):
            if response_lines[i].startswith("HTTP"):
                status_line = response_lines[i]
                response.status_code = int(status_line.split()[1])
                break
            
        # exit_on_fail
        if exit_on_fail and not (200 <= response.status_code < 300):
            msg = 'Request to server failed'
            try:
                data = json.loads(response.stdout)
                if 'messages' in data and type(data['messages'] == list):
                    msg += "\nServer message: " + str(data['messages'][0])
            except:
                data = {}
                msg += ('\n' + "Couldn't parse server response")

            cprint(LRED, msg)
            if 'messages' in data and data['messages'][0] == 'invalid token':
                print("---")
                print("Use `tst login` to log in to the server")

            sys.exit(1)
        
        response.body = stdout if response.status_code else None
        
        return response


    def get(self, path, headers={}, exit_on_fail=False):
        return self.request('get', path, headers, exit_on_fail=exit_on_fail)


    def post(self, path, headers={}, payload='', exit_on_fail=False):
        return self.request('post', path, headers=headers, payload=payload, exit_on_fail=exit_on_fail)


    def patch(self, path, payload, headers={}, exit_on_fail=False):
        return self.request('patch', path, headers=headers, payload=payload, exit_on_fail=exit_on_fail)


    def delete(self, path, payload='', headers={}, exit_on_fail=False):
        return self.request('delete', path, headers=headers, payload=payload, exit_on_fail=exit_on_fail)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import tst
import sys

from tst.colors import *
from tst.utils import cprint
from tst.jsonfile import JsonFile

def main():
    ls(sys.argv)


def ls(args):
    """list files in activity"""

    tstjson = JsonFile('.tst/tst.json')
    files = tstjson.setdefault('files', {})
    for fn in sorted(files.keys()):
        visibility = files[fn].get('category', 'private')
        print(fn, ("(%s)" % visibility if visibility == 'public' else ''))



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import input
from uuid import getnode as get_mac

import sys
import os
import webbrowser
import logging

from requests import ConnectionError

import tst
from tst.colors import *
from tst.utils import cprint, data2json, _assert
from tst.jsonfile import JsonFile

def main():
    try:
        sitename = sys.argv[2] if len(sys.argv) > 2 else '_DEFAULT'
        login(sitename)
    except Exception as e:
        cprint(LRED, "ops!")
        raise e


def login_old(sitename):
    """login in site"""

    # fetch site urls
    site = tst.get_site(sitename)
    _assert(site is not None, "Site %s not found in config.yaml" % sitename)

    login_url = site.login_url()
    token_url = site.token_url()
    _assert(login_url, "Site %s has no login url" % site.name)
    _assert(token_url, "Site %s has no token url" % site.name)

    # open login url to user
    cprint(LGREEN, "Get a code at: %s" % login_url)
    webbrowser.open(login_url)
    code = input(LCYAN + "Code? " + RESET)

    # exchange code for token
    cprint(RESET, "Validating code: %s" % code)

    try:
        response = site.post(token_url, data={"code": code})
        _assert(response.ok, "Login failed")
    except ConnectionError:
        _assert(False, "Connection failed... check your internet connection")

    response.encoding = 'utf-8'
    try:
        token = response.json()
    except ValueError:
        _assert(False, "Server didn't send json")

    # save token
    tokens = JsonFile(os.path.expanduser('~/.tst/tokens.json'))
    tokens[site.name] = token['tst_token']
    tokens.writable = True
    tokens.save()

    msg = "Logged in %s as %s" % (site.name, token['email'])
    cprint(LGREEN, msg)
    logging.info(msg)


def login(sitename):
    """login in site"""

    # fetch site urls
    site = tst.get_site(sitename)
    _assert(site is not None, "Site %s not found in config.yaml" % sitename)

    # check site login style
    api_login_url = site.api_login_url()
    old_login_url = site.login_url();
    _assert(api_login_url or old_login_url, "Site %s has no login url" % site.name)

    # possibly switch to old style login function
    if not api_login_url and old_login_url:
        return login_old(sitename)

    # perform cli tools login; fetch auth and acces urls and login token
    response = site.post(api_login_url, {"mac": str(get_mac())})
    response.encoding = 'utf-8'
    user_auth_url = response.json()['user-auth-url']
    api_access_url = response.json()['api-access-url']

    # open browser at authorization page
    cprint(LCYAN, f"Open up the browser to authorize login: (Y/n) ", end="")
    if input() == "":
        print(f"Waiting for your authorization…")
        if not webbrowser.open(user_auth_url):
            cprint(LYELLOW, "Sorry, I cannot open the browser, visit:")
            cprint(LYELLOW, user_auth_url)
    else:
        print(f"Then, visit the url below:\n{user_auth_url}")

    # request access code
    for i in range(10):
        response = site.get(f'{api_access_url}')
        if response.status_code == 500:
            cprint(LRED, "Login failed (timeout)")
            cprint(LGREEN, "Trying again...")
            #sys.exit(1)
            continue
        authorization = response.json()
        if response.status_code == 200: break

    # check authorization
    #cprint(YELLOW, authorization)
    if 'authorized' not in authorization:
        cprint(LRED, 'Login not authorized')
        logging.info('login not authorized')
        sys.exit(1)

    # check whether the login worked
    if not 'tst_token' in authorization:
        cprint(LRED, 'Login failed (not authorized)')
        logging.info('no tst_token in authorized login')
        return

    # save token
    tokens = JsonFile(os.path.expanduser('~/.tst/tokens.json'))
    tokens[site.name] = authorization['tst_token']
    tokens.writable = True
    tokens.save()

    msg = f"Logged in {site.name} as {YELLOW}{authorization['email']}{RESET}"
    cprint(LGREEN, msg)
    logging.info(msg)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

YELLOW = '\033[1;33m'
LRED = '\033[1;31m'
LGREEN = '\033[1;32m'
GREEN = '\033[9;32m'
WHITE = '\033[0;37m'
LWHITE = '\033[1;37m'
LCYAN = '\033[1;36m'
LBLUE = '\033[1;34m'
RESET = '\033[0m'
CRITICAL = '\033[41;37m'



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .utils import _assert, cprint
import tst.popargs
from .colors import *

from .tst import save_assignment
from .tst import coverit
from .tst import get_config
from .tst import dirtype
from .tst import read_specification
from .tst import get_site
from .tst import save_files



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

def pop_argument(args, index=0):
    if index >= len(args):
        return None
    
    return args.pop(index)


def pop_option(args, option, short=None, default=None, vtype=str):
    if type(option) == list:
        selectors = ['--' + opt for opt in option]
    else:
        selectors = ['--' + option]

    if short is not None:
        selectors.append('-' + short)

    index = next((i for i in range(len(args)) if args[i] in selectors), None)
    if index is None or index == len(args) - 1:
        return default

    value = args.pop(index + 1)
    selector = args.pop(index)
    
    return vtype(value)


def pop_flag(args, flag, short=None):
    selectors = ['--' + flag]
    if short is not None:
        selectors.append('-' + short)

    index = next((i for i in range(len(args)) if args[i] in selectors), None)
    if index is None:
        return False

    args.pop(index)
    return True





from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import sys
import requests
import pkg_resources
from cachecontrol import CacheControl
from cachecontrol.caches.file_cache import FileCache

from tst.utils import cprint
from tst.colors import *

def main():
    current = pkg_resources.get_distribution('tst').version
    if not sys.stdout.isatty():
        print(current)
        return

    cprint(WHITE, current, file=sys.stdout)
    try:
        s = requests.session()
        s = CacheControl(s, cache=FileCache(os.path.expanduser('~/.tst/cache')))
        response = s.get('https://pypi.org/pypi/tst/json', timeout=5)
        data = response.json()

        latest_version = data['info']['version']
        if current != latest_version:
            cprint(YELLOW, 'Latest stable version: %s' % latest_version, file=sys.stdout)
            cprint(RESET, '---\n'
                          'Use `pip install --upgrade --user tst` to install')
    except requests.ConnectionError:
        cprint(LRED, 'Cannot connect to pypi.org server')



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import tst
import sys

from tst.colors import *
from tst.utils import cprint

def main():
    cprint(LGREEN, tst.dirtype() or "no tst directory")
    print("Script executed: " + sys.argv[0])



from setuptools import setup, find_packages
import codecs

setup(name='tst',
      version="0.12.0",
      description='TST Student Testing',
      url='http://github.com/daltonserey/tst',
      author='Dalton Serey',
      author_email='daltonserey@gmail.com',
      license='MIT',
      packages=find_packages(),
      include_package_data=True,
      scripts=[
        'bin/runjava',
        'tst/commands/tst-test',
      ],
      python_requires='>=3.6',
      install_requires=[
        'pyyaml>=5.1',
        'requests>=2.6.1',
        'cachecontrol[filecache]'
      ],
      entry_points = {
        'console_scripts': [
            'tst=tst.commands:main',
        ]
      },
      zip_safe=False)



# jsonfile
# (C) Dalton Serey - UFCG

"""
# usage

Use it as a simple dict with a save method that saves the data to the file. If
the file exists it will be read at instantiation. If you create more than one
object for the same filename, the same object will be returned by the
constructor (it is a singleton). See the example below.

```
from jsonfile import JsonFile

f = JsonFile('/Users/dalton/somefile.json', writable=True)
f['name'] = 'dalton'
f['data'] = [1, 2, 3]
f.save()
f2 = JsonFile('/Users/dalton/somefile.json')
f2 is f # True
```
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import str

import io
import sys
import os
import json

def to_unicode(obj, encoding='utf-8'):
    # 2to3: assert isinstance(obj, basestring), type(obj)
    assert isinstance(obj, str), type(obj)
    if isinstance(obj, str):
        return obj

    for encoding in ['utf-8', 'latin1']:
        try:
            obj = str(obj, encoding)
            return obj
        except UnicodeDecodeError:
            pass

    assert False, "jsonfile: non-recognized encoding"

class CorruptedJsonFile(Exception): pass

DEFAULT_FAIL_MESSAGE = "file is corrupted"

class JsonFile(object):

    __instances = {}

    def __new__(cls, filename, exit_on_fail=False, failmsg=None, array2map=None, writable=False):
        filename = os.path.expanduser(filename)
        if filename in JsonFile.__instances:
            return JsonFile.__instances[filename]

        JsonFile.__instances[filename] = object.__new__(cls)
        self = JsonFile.__instances[filename]
        self.filename = filename
        self.writable = writable
        self.isjson = filename.endswith("json")
        self.isyaml = filename.endswith("yaml") or filename.endswith("yml")
        
        if os.path.exists(filename):
            self.load(failmsg=failmsg, exit_on_fail=exit_on_fail)
            if array2map and type(self.data) is list:
                self.data = { array2map: self.data }

        else:
            self.data = {}

        return self


    def __setitem__(self, key, value):
        self.data[key] = value


    def __getitem__(self, key):
        return self.data[key]


    def __contains__(self, key):
        return key in self.data


    def pop(self, key):
        self.data.pop(key, None)


    def setdefault(self, key, value):
        return self.data.setdefault(key, value)


    def load(self, exit_on_fail=False, failmsg=None):
        if not os.path.exists(self.filename):
            self.data = {}
            return

        # actually read data from file system
        if self.isjson:
            try:
                with io.open(self.filename, mode='r', encoding='utf-8') as f:
                    self.data = json.loads(to_unicode(f.read()))

            except ValueError as e:
                if exit_on_fail or failmsg:
                    print(failmsg or DEFAULT_FAIL_MESSAGE, file=sys.stderr)
                    sys.exit(1)

                raise CorruptedJsonFile("corrupted specification file")

        else:
            import yaml
            try:
                with io.open(self.filename, mode='r', encoding='utf-8') as f:
                    self.data = yaml.load(to_unicode(f.read()), Loader=yaml.FullLoader)
                    if self.data is None:
                        raise ValueError()

            except (ValueError, yaml.scanner.ScannerError) as e:
                if exit_on_fail or failmsg:
                    print(failmsg or DEFAULT_FAIL_MESSAGE, file=sys.stderr)
                    sys.exit()

                raise CorruptedJsonFile("unrecognized file")



    def save(self):
        assert self.writable, "jsonfile: cannot save a non writable JsonFile"
        with io.open(self.filename, mode="w", encoding='utf-8') as f:
            f.write(json.dumps(
                self.data,
                indent=2,
                separators=(',', ': ')
            ))


    def get(self, key, default=None):
        return self.data.get(key, default)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import sys
import logging
from subprocess import check_call, CalledProcessError

import tst
from tst.colors import *
from tst.utils import cprint

log = logging.getLogger('dispatcher')
handler_file = logging.FileHandler(os.path.expanduser('~/.tst/logs'))
handler_file.setFormatter(logging.Formatter('%(asctime)s|%(name)s|%(levelname)s|%(message)s'))
log.addHandler(handler_file)
log.setLevel(logging.DEBUG)

EXTERNALS = [
    "test",
    "status",
    "new",
    "delete",
]
DEFAULT_COMMAND = "test"


def run_external_command(command, args):
    script_name = os.path.expanduser("tst-%s" % command)
    args.insert(0, script_name)
    try:
        exit = check_call(args)
    except CalledProcessError as e:
        sys.exit(e.returncode)
        pass
    except OSError:
        print("tst: couldn't run command '%s'" % command, file=sys.stderr)


def identify_and_run_command(args):
    from distutils.spawn import find_executable
    if args and (args[0] in EXTERNALS):
        command_name = args.pop(0)
        run_external_command(command_name, args)

    elif args and find_executable('tst-%s' % args[0]):
        command_name = args.pop(0)
        run_external_command(command_name, args)
            
    else: # neither internal, nor external command!?
        command_name = DEFAULT_COMMAND
        run_external_command(command_name, args)

    return command_name


def dispatcher(args):
    possible_command = args[0] if args else None
    if possible_command in ['--version', '-v', 'version']:
        import tst.commands.version as version
        version.main()

    elif possible_command == 'info':
        import tst.commands.info as info
        info.main()

    elif possible_command == 'login':
        import tst.commands.login as login
        login.main()

    elif possible_command == 'ls':
        import tst.commands.ls as ls
        ls.main()

    elif possible_command == 'commit':
        dirtype = tst.dirtype()
        if dirtype == "old:activity":
            cprint(LRED, "┌──────────────────────────────────────────────────────────┐")
            cprint(LRED, "│ Oops...                                                  │")
            cprint(LRED, "│                                                          │")
            cprint(LRED, "│ This directory contains an activity in an old format. It │")
            cprint(LRED, "│ is no longer supported, however. Consider discarding     │")
            cprint(LRED, "│ this directory and performing a new checkout of the same │")
            cprint(LRED, "│ activity from the source site to have it updated to the  │")
            cprint(LRED, "│ latest version and current format.                       │")
            cprint(LRED, "└──────────────────────────────────────────────────────────┘")
            sys.exit(1)

        elif dirtype in ["assignment", None]:
            import tst.commands.commit as commit
            commit.main()

        else:
            cprint(LRED, "No assignment found in this directory")
            sys.exit(1)

    elif possible_command == 'checkout':
        dirtype = tst.dirtype()
        if dirtype == "old:activity":
            cprint(YELLOW, "┌─────────────────────────────────────────────────────────┐")
            cprint(YELLOW, "│ IMPORTANT                                               │")
            cprint(YELLOW, "│                                                         │")
            cprint(YELLOW, "│ This is an old style activity. The old checkout command │")
            cprint(YELLOW, "│ will be used.                                           │")
            cprint(YELLOW, "└─────────────────────────────────────────────────────────┘")
            run_external_command("checkout", args[1:])

        elif dirtype not in [None, "assignment"]:
            cprint(LRED, "┌───────────────────────────────────────────────┐")
            cprint(LRED, "│ checkout cannot be executed in this directory │")
            cprint(LRED, "│                                               │")
            cprint(LRED, "│ Currently, the checkout command must be used  │")
            cprint(LRED, "│ in a non tst directory.                       │")
            cprint(LRED, "└───────────────────────────────────────────────┘")

        else:
            import tst.commands.checkout as checkout
            checkout.main()

    elif possible_command == 'update':
        cprint(LRED, "┌─────────────────────────────────────┐")
        cprint(LRED, "│ update is deprecated                │")
        cprint(LRED, "│                                     │")
        cprint(LRED, "│ Use pip to update tst:              │")
        cprint(LRED, "│ $ pip install tst --upgrade --user  │")
        cprint(LRED, "└─────────────────────────────────────┘")

    elif possible_command == 'config':
        cprint(LRED, "┌────────────────────────────────────────────────────┐")
        cprint(LRED, "│ config is deprecated                               │")
        cprint(LRED, "│                                                    │")
        cprint(LRED, "│ Edit ~/.tst/config.yaml directly to configure tst. │")
        cprint(LRED, "│ See documentation to see existing options.         │")
        cprint(LRED, "└────────────────────────────────────────────────────┘")

    else:
        command = identify_and_run_command(args)


def main():
    try:
        args = sys.argv[:]
        args.pop(0) # pop dispatcher name
        config = tst.get_config()
        dispatcher(args) 
    except KeyboardInterrupt:
        cprint(LRED, "\nUser interruption")
